package org.embulk.parser.header_based_csv

import org.embulk.config._
import org.embulk.parser.header_based_csv.exception.{ColumnValueNotFoundException, TooManyColumnValuesException}
import org.embulk.parser.header_based_csv.output.ColumnValueWriter
import org.embulk.parser.header_based_csv.tokenizer.TokenizerHelper.RichCsvTokenizer
import org.embulk.spi.util.LineDecoder
import org.embulk.spi._
import org.embulk.standards.{CsvParserPlugin, CsvTokenizer}
import org.slf4j.Logger
import helper.GoogleOptionalHelper.RichOptional

import scala.annotation.tailrec
import scala.collection.JavaConverters._
import scala.util.{Failure, Success, Try}

class HeaderBasedCsvParserPlugin extends CsvParserPlugin {

  trait PluginTask extends CsvParserPlugin.PluginTask

  private val log: Logger = Exec.getLogger(classOf[CsvParserPlugin])

  override def transaction(config: ConfigSource, control: ParserPlugin.Control): Unit = {
    val task = config.loadConfig(classOf[PluginTask])

    val skipHeaderLines = task.getSkipHeaderLines

    // CsvParserPlugin had header_line option (not anymore)
    // super class has backward compatibility code in its "transaction" method and it discards the header line when header-line option is true
    // so you have to override the method and keep the header line
    task.getHeaderLine.toOption match {
      case Some(_) if skipHeaderLines > 0 =>
        throw new ConfigException("'header_line' option is invalid if 'skip_header_lines' is set.")
      case Some(hasHeader) =>
        if (hasHeader)
          task.setSkipHeaderLines(0) // keep the header line (should set 0 not to skip the first line unlike super class)
        else
          throw new ConfigException("this parser needs a header line")
      case None => task.setSkipHeaderLines(0)
    }

    control.run(task.dump, task.getSchemaConfig.toSchema)
  }

  override def run(taskSource: TaskSource, schema: Schema, input: FileInput, output: PageOutput): Unit =
    try {
      val task = taskSource.loadTask(classOf[PluginTask])

      val allowExtraColumnValues = task.getAllowExtraColumns
      val allowOptionalColumns   = task.getAllowOptionalColumns
      val stopOnInvalidRecord    = task.getStopOnInvalidRecord

      val skipHeaderLines = task.getSkipHeaderLines

      val tokenizer         = new CsvTokenizer(new LineDecoder(input, task), task)
      val pageBuilder       = new PageBuilder(Exec.getBufferAllocator, schema, output)
      val columnValueWriter = new ColumnValueWriter(task, pageBuilder)
      val columns = schema.getColumns.asScala.toList

      while (tokenizer.nextFile()) {

        (1 to skipHeaderLines).foreach(_ => tokenizer.skipHeaderLine())

        tokenizer.forEachRecord { record =>
          val result = writeColumnValues(columns,
                                         record,
                                         tokenizer,
                                         columnValueWriter,
                                         allowExtraColumnValues,
                                         allowOptionalColumns)
          skipOrStopIfFailed(result, tokenizer, stopOnInvalidRecord)
          if (result.isSuccess) {
            pageBuilder.addRecord()
          }
        }
        pageBuilder.flush()
      }
      pageBuilder.finish()
    } catch {
      case e: Throwable => e.printStackTrace()
    }

  private def writeSingleColumnValue(record: Map[String, String],
                                     column: Column,
                                     columnValueWriter: ColumnValueWriter,
                                     columnCount: Int,
                                     allowOptionalColumns: Boolean): Try[Unit] =
    record.get(column.getName) match {
      case Some(value) =>
        columnValueWriter.writeValue(column, value)

      case None if allowOptionalColumns && record.size < columnCount =>
        columnValueWriter.writeNull(column)

      case None =>
        Failure(new ColumnValueNotFoundException(column))
    }

  private def writeColumnValues(columns: Seq[Column],
                                record: Map[String, String],
                                tokenizer: CsvTokenizer,
                                columnValueWriter: ColumnValueWriter,
                                allowExtraColumnValues: Boolean,
                                allowOptionalColumns: Boolean): Try[Unit] = {

    @tailrec def loop(cs: Seq[Column]): Try[Unit] =
      cs match {
        case Nil =>
          Success(())

        // allowExtraColumns (which renamed to allowExtraColumnValues here because its more suitable)
        // "If true, ignore too many columns. Otherwise, skip the row in case of too many columns"
        case _ :: _ if !allowExtraColumnValues && columns.size < record.size =>
          Failure(new TooManyColumnValuesException(columns.size, record.size))

        case column :: tail =>
          writeSingleColumnValue(record, column, columnValueWriter, columns.length, allowOptionalColumns) match {
            case Success(_)    => loop(tail)
            case f: Failure[_] => f
          }
      }
    loop(columns)
  }

  private def skipOrStopIfFailed(result: Try[Unit], tokenizer: CsvTokenizer, stopOnInvalidRecord: Boolean): Unit =
    result.failed
      .foreach { e =>
        val skippedLine = tokenizer.skipCurrentLine()
        val lineNumber  = tokenizer.getCurrentLineNumber
        if (stopOnInvalidRecord)
          throw new DataException(s"Invalid record at line $lineNumber: $skippedLine", e)
        else
          log.warn(s"Skipped line $lineNumber (${e.getMessage}): $skippedLine")
      }

}
